/* 
 * File:   main.c
 * Author: gallardice
 *
 * Template for CSUB LCDPANEL Drivers
 *
 *
 */


/*
arm-none-eabi-gcc -c -mcpu=cortex-m4 -mfpu=fpv4-sp-d16 -mfloat-abi=hard -mthumb -Wall -Wno-main -ffunction-sections -O0 -fPIC -I. main.c -o main.o
arm-none-eabi-ld -nostartfiles -T arm-gcc-link.ld -o main.elf main.o


*/
/* Uncomment the relevant  #define statement to use the required code   */
#define ST7789
//#define QUICKSORT
//#define STRREV

/**********************************************/

#include "../inc/ARMF4CFunctions.h"

#ifdef ST7789

    #ifdef EXAMPLEMMBASICCODE
            '****** MMBASIC Sample program to call this CSUB ************'
            ' ST7789 dc,rst,cs,orientation'
              ST7789 64,63,58,1                           'Call the CSUB'
			  CLS RGB(BLUE)                               'Rest of program'
              PRINT "HELLO"
			  CSUB ST7789 INTEGER,INTEGER,INTEGER,INTEGER 'CSUB definition with optional types '
			  ...
			  END CSUB
    #endif

/*******************************************************************************
 *
 * Driver for ST7789 Display written as CSUB for ARMMITE F4
 *
 * Acknowledgements to Peter Mather & Geoff Graham & Peter Carnegie
 * for previous work this is derived from.
 *
 *
 * When Generating the CSUB, use MERGE mode, and name the SUB
 * ST7789
 *
 * Entry point is CSUB
 * void main(long long *DC,long long *RST,long long *CS,long long *orientation)
 * e.g.       ST7789 64,63,58,1               // PC7,PC6,PD11
 *                                      1 is LANDSCAPE
 *                                      2 is PORTRAIT
 *                                      3 is RLANDSCAPE
 *                                      4 is RPORTRAIT
 *
 * V1.0     2021-10-08 Gerry Allardice For ST7789 1.3" IPS 240x240
 *          bitbanged using the SPI2 pins. For ARMMiteF4
 *
 *
 ******************************************************************************/


#define Version 100     //Version 1.00
#include <stdarg.h>

#define ST7735_NOP              0x0
#define ST7735_SWRESET          0x01
#define ST7735_RDDID            0x04
#define ST7735_RDDST            0x09
#define ST7735_SLPIN            0x10
#define ST7735_SLPOUT           0x11
#define ST7735_PTLON            0x12
#define ST7735_NORON            0x13
#define ST7735_INVOFF           0x20
#define ST7735_INVON            0x21
#define ST7735_DISPOFF          0x28
#define ST7735_DISPON           0x29
#define ST7735_CASET            0x2A
#define ST7735_RASET            0x2B
#define ST7735_RAMWR            0x2C
#define ST7735_RAMRD            0x2E
#define ST7735_PTLAR            0x30
#define ST7735_MADCTL           0x36
#define ST7735_COLMOD           0x3A
#define ST7735_FRMCTR1          0xB1
#define ST7735_FRMCTR2          0xB2
#define ST7735_FRMCTR3          0xB3
#define ST7735_INVCTR           0xB4
#define ST7735_DISSET5          0xB6
#define ST7735_PWCTR1           0xC0
#define ST7735_PWCTR2           0xC1
#define ST7735_PWCTR3           0xC2
#define ST7735_PWCTR4           0xC3
#define ST7735_PWCTR5           0xC4
#define ST7735_VMCTR1           0xC5
#define ST7735_RDID1            0xDA
#define ST7735_RDID2            0xDB
#define ST7735_RDID3            0xDC
#define ST7735_RDID4            0xDD
#define ST7735_PWCTR6           0xFC
#define ST7735_GMCTRP1          0xE0
#define ST7735_GMCTRN1          0xE1
#define ST7735_Portrait         0x02
#define ST7735_Portrait180      0xc0
#define ST7735_Landscape        0x62
#define ST7735_Landscape180     0xA0
#define ST77XX_SWRESET    0x01
#define ST77XX_DISPON     0x29
#define ST77XX_CASET      0x2A
#define ST77XX_RASET      0x2B
#define ST77XX_INVON      0x21
#define ST77XX_NORON      0x13
#define ST77XX_SLPOUT     0x11
#define ST77XX_COLMOD     0x3A

#define LANDSCAPE       1
#define PORTRAIT        2
#define RLANDSCAPE      3
#define RPORTRAIT       4


#define SPI2
#ifdef SPI2
 // SPI2 pin numbers for ARMMITE F4  (LCD)
 #define SPI_CLK_PIN        52
 #define SPI_INP_PIN        53
 #define SPI_OUT_PIN        54
#else

 // SPI1 pin numbers for ARMMITE F4  W25Q16 FLASH
 #define SPI_CLK_PIN        89
 #define SPI_INP_PIN        90
 #define SPI_OUT_PIN        91
#endif


#define SPIsend(a) {SPI_send_byte(a);}


/* *************************************************
 * Optional dummy function to document the generated CSUB.
 * Uses only a few bytes but allows identification of
 * the CSUBs function/use. Up to 60 characters
 * *************************************************/

 void ST7789_LCD_IPS_240x240_1_3_INCH_v100(){
  NOP();   //Padding to make sure it ends on 4 byte boundary
 }

/* **********************************************************************
 * This function (getFPC) does NOT work and is also NOT required for ARMMITE F4
 * It is shown here to document this fact. Do NOT include it.
 * ********************************************************************/
 // __attribute__((noinline)) void getFPC(void *a, void *b, volatile unsigned int *c)
 //     {
 //        *c = (unsigned int) (__builtin_return_address (0) - (b -a) - 1) ;
 //     }

 void p_string(const char *s){
    char  * testData    = ( char *)((void *)s );
    MMPrintString(testData);
 }

 void p_int(int a,int base){
    char b[64];
    IntToStr(b,a,base);
    MMPrintString(b);putConsole(32);
 }

 /*
  * Simultaneously transmit and receive a byte on the SPI.
  *
  * Polarity and phase are assumed to be both 0, i.e.:
  *   - input data is captured on rising edge of SCLK.
  *   - output data is propagated on falling edge of SCLK.
  *
  * Returns the received byte.
  */

 void SPI_send_byte(unsigned int byte_out)
 {
     //unsigned char byte_in = 0;
     unsigned int bit;
     for (bit = 0x80; bit; bit >>= 1) {

    	 /* Shift-out a bit to the MOSI line */
         if(byte_out & bit){
           PinSetBit(SPI_OUT_PIN,LATSET);
         }else{
           PinSetBit(SPI_OUT_PIN,LATCLR);
         }

         /* Delay for at least the peer's setup time */
         NOP(); NOP();

         /* Pull the clock line high */
         PinSetBit(SPI_CLK_PIN,LATSET);

         /* Shift-in a bit from the MISO line */
         //if (PinRead(SPI_INP_PIN) == 0x1)
         //    byte_in |= bit;

         /* Delay for at least the peer's hold time */
          NOP(); NOP();

         /* Pull the clock line low */
         PinSetBit(SPI_CLK_PIN,LATCLR);
     }

     //return byte_in;
 }
void spi_write_command(unsigned int data){
      PinSetBit(Option.LCD_CD, LATCLR);
      PinSetBit(Option.LCD_CS, LATCLR);
      SPI_send_byte(data);
      PinSetBit(Option.LCD_CS, LATSET);
}

 void spi_write_data(unsigned int data){
      PinSetBit(Option.LCD_CD, LATSET);
      PinSetBit(Option.LCD_CS, LATCLR);
      SPI_send_byte(data);
      PinSetBit(Option.LCD_CS, LATSET);
}

void spi_write_cd(unsigned int command, unsigned int data, ...){
      int i;
      va_list ap;
      va_start(ap, data);
      spi_write_command(command);
       for(i = 0; i < data; i++) spi_write_data((char)va_arg(ap, int));
       va_end(ap);
}


/*******************************************************************************
* defines start/end coordinates for memory access from host to LCDPANEL
* also maps the start and end points to suit the orientation
*******************************************************************************/
void DefineRegion(int xstart, int ystart, int xend, int yend) {
    if(Option.DISPLAY_ORIENTATION==4){   //RP
        ystart+=80;
        yend+=80;
    }
    if(Option.DISPLAY_ORIENTATION==3){  //RL
        xstart+=80;
        xend+=80;
    }

    PinSetBit(Option.LCD_CD, LATCLR);
    PinSetBit(Option.LCD_CS, LATCLR);
    SPIsend(ST7735_CASET);
    PinSetBit(Option.LCD_CD, LATSET);
    SPIsend(xstart >> 8 );
    SPIsend(xstart );
    SPIsend(xend >> 8 );
    SPIsend(xend );
    PinSetBit(Option.LCD_CD, LATCLR);
    SPIsend(ST7735_RASET);
    PinSetBit(Option.LCD_CD, LATSET);
    SPIsend(ystart >> 8 );
    SPIsend(ystart );
    SPIsend(yend >> 8 );
    SPIsend(yend );
    PinSetBit(Option.LCD_CD, LATCLR);
    SPIsend(ST7735_RAMWR);
    PinSetBit(Option.LCD_CS, LATSET);

}



/********************************************************************
*  DrawRectangle in MMBasic is redirected to this function
* this is the basic drawing primitive used by most drawing routines
*    x1, y1, x2, y2 - the coordinates
*    c - the colour
* ------------------------------------------------------
* ***** MUST BE static void for the ARMMITEF4 **********
* ------------------------------------------------------
* *********************************************************************/
static void DrawRectangleSPI(int x1, int y1, int x2, int y2, int c){
        int i, t;
        unsigned char hb, lb;

         // make sure the coordinates are kept within the display area
         if(x2 <= x1) { t = x1; x1 = x2; x2 = t; }
         if(y2 <= y1) { t = y1; y1 = y2; y2 = t; }
         if(x1 < 0) x1 = 0;
         if(x1 >= HRes) x1 = HRes - 1;
         if(x2 < 0) x2 = 0;
         if(x2 >= HRes) x2 = HRes - 1;
         if(y1 < 0) y1 = 0;
         if(y1 >= VRes) y1 = VRes - 1;
         if(y2 < 0) y2 = 0;
         if(y2 >= VRes) y2 = VRes - 1;

         // convert the colours to 565 format
         hb = ((c >> 16) & 0b11111000) | ((c >> 13) & 0b00000111);
         lb = ((c >> 5) & 0b11100000) | ((c >> 3) & 0b00011111);

         DefineRegion(x1, y1, x2, y2);
         PinSetBit(Option.LCD_CD, LATSET);                               //set CD high
         PinSetBit(Option.LCD_CS, LATCLR);
     	i = x2 - x1 + 1;
     	i *= (y2 - y1 + 1);

         while(i--){
            SPIsend(hb);
            SPIsend(lb);
         }

         PinSetBit(Option.LCD_CS, LATSET);

}


/********************************************************************
* DrawBitMap in MMBasic is redirected to this function
*   Prints the bitmap of a char
*    x, y - the top left of the char
*    width, height - size of the char's bitmap
*    scale - how much to scale the bitmap
* 	  fc, bc - foreground and background colour
*   bitmap - pointer to the bitmap
* ------------------------------------------------------
* ***** MUST BE static void for the ARMMITEF4 **********
* ------------------------------------------------------
* **********************************************************************/
static void DrawBitmapSPI(int x1, int y1, int width, int height, int scale, int fc, int bc, unsigned char *bitmap){
         int i, j, k, m;
         unsigned char fhb, flb, bhb, blb;

         int vertCoord, horizCoord, XStart, XEnd, YEnd;
         if(x1>=HRes || y1>=VRes || x1+width*scale<0 || y1+height*scale<0)return;

         // adjust when part of the bitmap is outside the displayable coordinates
         vertCoord = y1; if(y1 < 0) y1 = 0;                                 // the y coord is above the top of the screen
         XStart = x1; if(XStart < 0) XStart = 0;                            // the x coord is to the left of the left marginn
         XEnd = x1 + (width * scale) - 1; if(XEnd >= HRes) XEnd = HRes - 1; // the width of the bitmap will extend beyond the right margin
         YEnd = y1 + (height * scale) - 1; if(YEnd >= VRes) YEnd = VRes - 1;// the height of the bitmap will extend beyond the bottom margin

          // convert the colours to 565 format
         fhb = ((fc >> 16) & 0b11111000) | ((fc >> 13) & 0b00000111);
         flb = ((fc >>  5) & 0b11100000) | ((fc >>  3) & 0b00011111);
         bhb = ((bc >> 16) & 0b11111000) | ((bc >> 13) & 0b00000111);
         blb = ((bc >>  5) & 0b11100000) | ((bc >>  3) & 0b00011111);

         DefineRegion(x1, y1, x1 + (width * scale) - 1, y1 + (height * scale) -1);
         PinSetBit(Option.LCD_CD, LATSET);                               //set CD high
         PinSetBit(Option.LCD_CS, LATCLR);
         for(i = 0; i < height; i++) {                                   // step thru the font scan line by line
             for(j = 0; j < scale; j++) {                                // repeat lines to scale the font
                 if(vertCoord++ < 0) continue;                           // we are above the top of the screen
                 if(vertCoord > VRes) return;                            // we have extended beyond the bottom of the screen
                 horizCoord = x1;
                 for(k = 0; k < width; k++) {                            // step through each bit in a scan line
                     for(m = 0; m < scale; m++) {                        // repeat pixels to scale in the x axis
                         if(horizCoord++ < 0) continue;                  // we have not reached the left margin
                         if(horizCoord > HRes) continue;                 // we are beyond the right margin
                         if((bitmap[((i * width) + k)/8] >> (((height * width) - ((i * width) + k) - 1) %8)) & 1) {
                             SPIsend(fhb );SPIsend(flb );
                         } else {
                             SPIsend(bhb); SPIsend(blb);
                         }
                     }
                 }
             }
         }

         PinSetBit(Option.LCD_CS, LATSET);
}


/*******************************************************************************
 *
 * ST7789 : Initialise the CSUB Driver Sub-System
 *
 * CSUB called to initialise the driver SubSystem
 *
 * ST7789 is ALWAYS called from an MMBasic program
 * On exit, vectors DrawRectangleVector, and DrawBitmapVector will
 * be set to point to the CSUBs DrawRectangleSPI and
 * DrawBitmapSPI respectively
 *
 * Entry point is csub main(long long *DC,
 *                          long long *RST,
 *                          long long *CS
 *                          long long *orientation)
 *e.g. ST7789 64,63,58,1
 *
 ******************************************************************************/

void main(long long *CD, long long *RST, long long *CS,long long *orientation){
    int HorizontalRes=240,VerticalRes=240;

    Option.LCD_Reset=*RST;
    Option.LCD_CD=*CD;
    Option.LCD_CS=*CS;
    Option.DISPLAY_ORIENTATION=*orientation;
    Option.DISPLAY_TYPE=USERLCDPANEL ;

    if(ExtCurrentConfig[(int) Option.LCD_Reset]!=EXT_BOOT_RESERVED){ExtCfg(Option.LCD_Reset,EXT_DIG_OUT,0);}
    //if(ExtCurrentConfig[(int) Option.LCD_Reset]!=EXT_BOOT_RESERVED){ExtCfg(Option.LCD_Reset,EXT_DIG_OUT,0);ExtCfg(Option.LCD_Reset,EXT_BOOT_RESERVED,0);}
    PinSetBit(Option.LCD_Reset, LATSET);

    if(ExtCurrentConfig[(int) Option.LCD_CD]!=EXT_BOOT_RESERVED){ExtCfg(Option.LCD_CD,EXT_DIG_OUT,0);}
    //if(ExtCurrentConfig[(int) Option.LCD_CD]!=EXT_BOOT_RESERVED){ExtCfg(Option.LCD_CD,EXT_DIG_OUT,0);ExtCfg(Option.LCD_CD,EXT_BOOT_RESERVED,0);}
    PinSetBit(Option.LCD_CD, LATSET);

    if(ExtCurrentConfig[(int) Option.LCD_CS]!=EXT_BOOT_RESERVED){ExtCfg(Option.LCD_CS,EXT_DIG_OUT,0);}
    //if(ExtCurrentConfig[(int) Option.LCD_CS]!=EXT_BOOT_RESERVED){ExtCfg(Option.LCD_CS,EXT_DIG_OUT,0);ExtCfg(Option.LCD_CS,EXT_BOOT_RESERVED,0);}
    PinSetBit(Option.LCD_CS, LATSET);


    if(ExtCurrentConfig[SPI_OUT_PIN]!=EXT_BOOT_RESERVED){
        ExtCfg(SPI_OUT_PIN, EXT_DIG_OUT, 0); //ExtCfg(SPI_OUT_PIN, EXT_BOOT_RESERVED, 0);
        ExtCfg(SPI_INP_PIN, EXT_DIG_IN, 0); //ExtCfg(SPI_INP_PIN, EXT_BOOT_RESERVED, 0);
        ExtCfg(SPI_CLK_PIN, EXT_DIG_OUT, 0);// ExtCfg(SPI_CLK_PIN, EXT_BOOT_RESERVED, 0);
       //Set pins for SPI Mode 0 to initial states
         PinSetBit(SPI_CLK_PIN,LATCLR);
         PinSetBit(SPI_OUT_PIN,LATCLR);
         PinSetBit(SPI_INP_PIN,LATCLR);
    }


   //Reset the ST7789
    PinSetBit(Option.LCD_Reset,LATSET); uSec(10000);
    PinSetBit(Option.LCD_Reset,LATCLR); uSec(100000);
    PinSetBit(Option.LCD_Reset,LATSET); uSec(200000);

    spi_write_command(ST77XX_SWRESET);uSec(150000);
    spi_write_command(ST77XX_SLPOUT);uSec(500000);

    spi_write_command(ST77XX_COLMOD);spi_write_data(0x55);uSec(10000);

    // spi_write_command(ST77XX_CASET); spi_write_data(0x0); spi_write_data(0x0); spi_write_data(0x0); spi_write_data(240);
   // spi_write_command(ST77XX_RASET); spi_write_data(0x0); spi_write_data(0); spi_write_data(0); spi_write_data(240);

   // spi_write_cd(ST7735_FRMCTR1,3,0x01,0x2C,0x2d);               //frame rate control - normal mode
	//spi_write_cd(ST7735_FRMCTR2,3,0x01,0x2C,0x2D);               //frame rate control - idle mode
	//spi_write_cd(ST7735_FRMCTR3,6,0x01,0x2c,0x2D,0x01,0x2C,0x2D); //frame rate control - partial mode

    //spi_write_cd(0xB2,5,0x0C,0x0C,0x00,0x33,0x33);
	//spi_write_cd(0xB7,1,0x35);
	//spi_write_cd(0xBB,1,0x19);
	//spi_write_cd(0xC0,1,0x2C);
	//spi_write_cd(0xC2,1,0x01);
	//spi_write_cd(0xC3,1,0x12);
	//spi_write_cd(0xC4,1,0x20);
	//spi_write_cd(0xC6,1,0x0F);
	//spi_write_cd(0xD0,2,0xA4,0xA1);
	//spi_write_cd(0xE0,14,0xD0,0x04,0x0D,0x11,0x13,0x2B,0x3F,0x54,0x4C,0x18,0x0D,0x0B,0x1F,0x23);
	//spi_write_cd(0xE1,14,0xD0,0x04,0x0C,0x11,0x13,0x2C,0x3F,0x44,0x51,0x2F,0x1F,0x1F,0x20,0x23);



    spi_write_command(ST77XX_INVON);    uSec(10000);
    spi_write_command(ST77XX_NORON);    uSec(10000);
    spi_write_command(ST77XX_DISPON);   uSec(500000);

    if(Option.DISPLAY_ORIENTATION==LANDSCAPE)     spi_write_cd(ST7735_MADCTL, 1, ST7735_Landscape  );
    if(Option.DISPLAY_ORIENTATION==PORTRAIT)      spi_write_cd(ST7735_MADCTL, 1, ST7735_Portrait  );
    if(Option.DISPLAY_ORIENTATION==RLANDSCAPE)    spi_write_cd(ST7735_MADCTL, 1, ST7735_Landscape180  );
    if(Option.DISPLAY_ORIENTATION==RPORTRAIT)     spi_write_cd(ST7735_MADCTL, 1, ST7735_Portrait180 );

    if(Option.DISPLAY_ORIENTATION&1){
        VRes=HorizontalRes;
        HRes=VerticalRes;
    } else {
        HRes=HorizontalRes;
        VRes=VerticalRes;
    }

    //update original Vectors to point to our new functions

    #ifdef BLITSUPPORTED
    //These allow BLIT and the function Pixel() to work if you have written the relevant functions.
    //You would need to be able to read from the display to implement ReadBufferSPI
    //LOAD IMAGE filename will use DrawRectangleSPI and write a pixel at a time if WriteBuffererSPI is not implemented
    //If WriteBuffererSPI is available it will use that and will be quicker as it only has to define the area to load once
    //instead of once per pixel.
    ReadBufferVector=(unsigned int)&ReadBufferSPI ;
    DrawBufferVector=(unsigned int)&DrawBufferSPI ;
    #endif

    //Set the DrawBitmap vector to point to our function
    // Called each time  the TEXT command writes a character
    DrawBitmapVector=(unsigned int)&DrawBitmapSPI;

    //Set the DrawRectangle vector to point to our function
    // Called when any of the GUI commands writes to screen.(Also CLS)
    DrawRectangleVector= (unsigned int)&DrawRectangleSPI;

    //CLS
    DrawRectangle(0,0,HRes-1,VRes-1,0x000000);

    static const char startup[]="ST7789 driver bitbanged SPI loaded\r\n";
    //static const char crlf[]="\r\n";
    //static const char s1[]="First variable = ";
    //static const char s2[]="Second variable = ";
    //static const char s3[]="Third variable = ";
    //static const char s4[]="First variable = ";
    // static const char s5[]="Second variable = ";
   // static const char s6[]="LCD = ";
    p_string(startup);
   // p_string(s1); p_int(Option.LCD_CD,10);   p_string(crlf);
   // p_string(s2);p_int(Option.LCD_Reset,10); p_string(crlf);
    //p_string(s3); p_int(Option.LCD_CS,10);   p_string(crlf);
   // p_string(s6); p_int(HRes,10);p_int(VRes,10);p_int(Option.DISPLAY_TYPE,10);p_int(Option.DefaultBrightness,10); p_string(crlf);
}

#endif



#ifdef STRREV
   #ifdef EXAMPLEMMBASICCODE
            '****** MMBASIC Sample program to call this CSUB ************'
              a$="ABCDEFG"
              STRREV a$,b$                           'Call the CSUB'
			  print b$
			  CSUB STRREV STRING,STRING   'CSUB definition with optional types '
			  ...
			  END CSUB
   #endif

/*
* This CFunction will reverse the characters in an MMBasic string
* eg if the string is "ABCDEFG" then on return the string will
* be "GFEDCBA"
*/

void strrev(unsigned char in[], unsigned char out[]) {

   unsigned int i;

   unsigned int len;

   //MMBasic stores the length of the string in the
   //first,0th,byte
   len=in[0];

   //Set the length of the output string
   out[0]=len;

   //Do the reverse copy
   for(i=1;i<(len+1);i++){
       out[i]=in[len-i+1];
   }


}

#endif

#ifdef QUICKSORT

// quicksort from rosetta stone
void quicksort(long long A[], long long *nElements) {
  long len=*nElements;
	if (len < 2) return;

  long long pivot = A[len / 2];

  long long i, j, h;
  for (i = 0, j = len - 1; ; i++, j--) {
    while (A[i] < pivot) i++;
    while (A[j] > pivot) j--;

    if (i >= j) break;

    long long temp = A[i];
    A[i]           = A[j];
    A[j]           = temp;
  }
  h=len-i;
  quicksort(A, &i);
  quicksort(A + i, &h);
}
#endif
